import datetime

import requests
from empiric.core.entry import construct_entry
from empiric.core.utils import currency_pair_to_key


def fetch_coingecko(assets, publisher):
    source = "coingecko"

    headers = {
        "Accepts": "application/json",
    }

    entries = []

    for asset in assets:
        if asset["type"] != "SPOT":
            print(f"Skipping Coingecko for non-spot asset {asset}")
            continue

        pair = asset["pair"]
        key = currency_pair_to_key(*pair)

        if pair[0] == "ETH":
            pair_id = "ethereum"
        elif pair[0] == "BTC":
            pair_id = "bitcoin"
        elif pair[0] == "SOL":
            pair_id = "solana"
        elif pair[0] == "AVAX":
            pair_id = "avalanche-2"
        elif pair[0] == "DOGE":
            pair_id = "dogecoin"
        elif pair[0] == "SHIB":
            pair_id = "shiba-inu"
        elif pair[0] == "TEMP":
            pair_id = "tempus"
        elif pair[0] == "DAI":
            pair_id = "dai"
        elif pair[0] == "USDT":
            pair_id = "tether"
        elif pair[0] == "USDC":
            pair_id = "usd-coin"
        elif pair[0] == "TUSD":
            pair_id = "true-usd"
        elif pair[0] == "BUSD":
            pair_id = "binance-usd"
        elif pair[0] == "BNB":
            pair_id = "binancecoin"
        elif pair[0] == "ADA":
            pair_id = "cardano"
        elif pair[0] == "XRP":
            pair_id = "ripple"
        elif pair[0] == "MATIC":
            pair_id = "matic-network"
        elif pair[0] == "AAVE":
            pair_id = "aave"
        else:
            raise Exception(
                f"Unknown price pair, do not know how to query coingecko for {pair[0]}"
            )

        url = f"https://api.coingecko.com/api/v3/coins/{pair_id}?localization=false&market_data=true&community_data=false&developer_data=false&sparkline=false"

        response = requests.get(url, headers=headers, timeout=10)
        response.raise_for_status()
        price = response.json()["market_data"]["current_price"][pair[1].lower()]
        timestamp = int(
            datetime.datetime.strptime(
                response.json()["last_updated"],
                "%Y-%m-%dT%H:%M:%S.%f%z",
            ).timestamp()
        )
        price_int = int(price * (10 ** asset["decimals"]))

        print(f"Fetched price {price} for {key} from Coingecko")

        entries.append(
            construct_entry(
                key=key,
                value=price_int,
                timestamp=timestamp,
                source=source,
                publisher=publisher,
            )
        )

    return entries



import requests
from empiric.core.entry import construct_entry
from empiric.core.utils import currency_pair_to_key


def fetch_bitstamp(assets, publisher):
    source = "bitstamp"
    base_url = "https://www.bitstamp.net/api/v2/ticker"

    entries = []

    for asset in assets:
        if asset["type"] != "SPOT":
            print(f"Skipping Bitstamp for non-spot asset {asset}")
            continue

        pair = asset["pair"]
        response = requests.get(
            f"{base_url}/{pair[0].lower()}{pair[1].lower()}", timeout=10
        )
        if response.status_code == 404:
            print(f"No data found for {'/'.join(pair)} from Bitstamp")
            continue

        result = response.json()

        timestamp = int(result["timestamp"])
        price = float(result["last"])
        price_int = int(price * (10 ** asset["decimals"]))
        key = currency_pair_to_key(*pair)

        print(f"Fetched price {price} for {'/'.join(pair)} from Bitstamp")

        entries.append(
            construct_entry(
                key=key,
                value=price_int,
                timestamp=timestamp,
                source=source,
                publisher=publisher,
            )
        )

    return entries



from .bitstamp import fetch_bitstamp
from .cex import fetch_cex
from .coinbase import fetch_coinbase
from .coingecko import fetch_coingecko
from .cryptowatch import fetch_cryptowatch
from .ftx import fetch_ftx
from .gemini import fetch_gemini
from .thegraph import fetch_thegraph



from empiric.core.base_client import EmpiricAccountClient, EmpiricBaseClient
from empiric.core.const import PUBLISHER_REGISTRY_ADDRESS
from empiric.core.entry import serialize_entries, serialize_entry


class EmpiricPublisherClient(EmpiricBaseClient):
    def __init__(
        self,
        publisher_private_key,
        publisher_address,
        publisher=None,
        publisher_registry_address=None,
        network=None,
        oracle_controller_address=None,
    ):
        self.publisher_registry_address = (
            publisher_registry_address
            if publisher_registry_address is not None
            else PUBLISHER_REGISTRY_ADDRESS
        )
        self.publisher = publisher
        super().__init__(
            publisher_private_key,
            publisher_address,
            network,
            oracle_controller_address,
        )
        # Override default account_client with one that uses timestamp for nonce
        self.account_client = EmpiricAccountClient(
            self.account_contract_address, self.client, self.signer
        )

    async def _fetch_contracts(self):
        await self._fetch_base_contracts()

    async def update_publisher_address(self, new_address, publisher=None):
        if self.publisher is None and publisher is None:
            raise ValueError(
                "No publisher provided at method call or instantiation, but need publisher ID to update address"
            )
        elif publisher is None:
            publisher = self.publisher

        result = await self.send_transaction(
            self.publisher_registry_address,
            "update_publisher_address",
            [publisher, new_address],
        )
        print(f"Updated publisher address with transaction {result}")

        return result

    async def publish(self, entry):
        result = await self.send_transaction(
            self.oracle_controller_address,
            "publish_entry",
            serialize_entry(entry),
        )
        print(f"Updated entry with transaction {result}")

        return result

    async def publish_many(self, entries):
        if len(entries) == 0:
            print("Skipping publishing as entries array is empty")
            return

        result = await self.send_transaction(
            self.oracle_controller_address,
            "publish_entries",
            serialize_entries(entries),
        )

        print(
            f"Successfully sent {len(entries)} updated entries with transaction {result}"
        )

        return result



from empiric.core.const import NETWORK, ORACLE_CONTROLLER_ADDRESS
from empiric.core.utils import str_to_felt
from starknet_py.contract import Contract
from starknet_py.net.gateway_client import GatewayClient


class EmpiricClient:
    def __init__(self, network=None, oracle_controller_address=None):
        if network is None:
            network = NETWORK
        if oracle_controller_address is None:
            oracle_controller_address = ORACLE_CONTROLLER_ADDRESS

        self.network = network
        self.oracle_controller_address = oracle_controller_address
        self.oracle_controller_contract = None

    async def fetch_oracle_controller_contract(self):
        if self.oracle_controller_contract is None:
            self.oracle_controller_contract = await Contract.from_address(
                self.oracle_controller_address,
                GatewayClient(self.network),
            )

    async def get_decimals(self, key):
        await self.fetch_oracle_controller_contract()

        if type(key) == str:
            key = str_to_felt(key)
        elif type(key) != int:
            raise AssertionError(
                "Key must be string (will be converted to felt) or integer"
            )

        response = await self.oracle_controller_contract.functions["get_decimals"].call(
            key
        )

        return response.decimals

    async def get_value(self, key, aggregation_mode, sources=None):
        await self.fetch_oracle_controller_contract()

        if type(key) == str:
            key = str_to_felt(key)
        elif type(key) != int:
            raise AssertionError(
                "Key must be string (will be converted to felt) or integer"
            )
        if sources is None:
            response = await self.oracle_controller_contract.functions[
                "get_value"
            ].call(key, aggregation_mode)
        else:
            response = await self.oracle_controller_contract.functions[
                "get_value_for_sources"
            ].call(key, aggregation_mode, sources)

        return (
            response.value,
            response.decimals,
            response.last_updated_timestamp,
            response.num_sources_aggregated,
        )

    async def get_entries(self, key, sources=None):
        await self.fetch_oracle_controller_contract()

        if type(key) == str:
            key = str_to_felt(key)
        elif type(key) != int:
            raise AssertionError(
                "Key must be string (will be converted to felt) or integer"
            )
        if sources is None:
            sources = []

        response = await self.oracle_controller_contract.functions["get_entries"].call(
            key, sources
        )

        return response.entries



import time

import requests
from empiric.core.entry import construct_entry


def fetch_thegraph(assets, publisher):
    source = "thegraph"

    base_url = "https://api.thegraph.com/subgraphs/name/"

    entries = []

    for asset in assets:
        if asset["type"] != "ONCHAIN":
            print(f"Skipping The Graph for non-on-chain asset {asset}")
            continue

        if asset["source"] == "AAVE":
            url_slug = "aave/protocol-v2"
            key = asset["key"]
            query = f"query {{reserves(where: {{id: \"{asset['detail']['asset_address']}\"}}) {{name isActive isFrozen {asset['detail']['metric']}}}}}"
            input_decimals = 27
        else:
            raise Exception(
                f"Unknown asset name, do not know how to query The Graph for {asset['name']}"
            )

        response = requests.post(base_url + url_slug, timeout=5, json={"query": query})
        result = response.json()["data"]["reserves"][0]

        assert result["name"] == asset["detail"]["asset_name"]
        assert result["isActive"] is True
        assert result["isFrozen"] is False

        value = float(result[asset["detail"]["metric"]])
        value_int = int(value * (10 ** (asset["decimals"] - input_decimals)))
        timestamp = int(time.time())

        print(f"Fetched data {value_int} for {key} from The Graph")

        entries.append(
            construct_entry(
                key=key,
                value=value_int,
                timestamp=timestamp,
                source=source,
                publisher=publisher,
            )
        )

    return entries



import time
from abc import ABC, abstractmethod

from empiric.core.const import NETWORK, ORACLE_CONTROLLER_ADDRESS
from starknet_py.contract import Contract
from starknet_py.net import AccountClient
from starknet_py.net.client_models import Call
from starknet_py.net.gateway_client import GatewayClient
from starknet_py.net.models import StarknetChainId
from starknet_py.net.networks import MAINNET, TESTNET
from starknet_py.net.signer.stark_curve_signer import KeyPair, StarkCurveSigner
from starkware.starknet.public.abi import get_selector_from_name


class EmpiricAccountClient(AccountClient):
    async def _get_nonce(self) -> int:
        return int(time.time())


class EmpiricBaseClient(ABC):
    def __init__(
        self,
        account_private_key,
        account_contract_address,
        network=None,
        oracle_controller_address=None,
    ):
        if network is None:
            network = NETWORK
        if oracle_controller_address is None:
            oracle_controller_address = ORACLE_CONTROLLER_ADDRESS

        if network == TESTNET:
            chain_id = StarknetChainId.TESTNET
        elif network == MAINNET:
            chain_id = StarknetChainId.MAINNET
        else:
            raise NotImplementedError(
                "Empiric.BaseClient: Network not recognized, unknown Chain ID"
            )

        self.network = network
        self.chain_id = chain_id
        self.oracle_controller_address = oracle_controller_address
        self.oracle_controller_contract = None
        self.account_contract_address = account_contract_address
        self.account_contract = None

        assert type(account_private_key) == int, "Account private key must be integer"
        self.account_private_key = account_private_key

        self.signer = StarkCurveSigner(
            self.account_contract_address,
            KeyPair.from_private_key(self.account_private_key),
            self.chain_id,
        )

        self.client = GatewayClient(self.network)
        self.account_client = AccountClient(
            self.account_contract_address, self.client, self.signer
        )

    @abstractmethod
    async def _fetch_contracts(self):
        pass

    async def _fetch_base_contracts(self):
        if self.oracle_controller_contract is None:
            self.oracle_controller_contract = await Contract.from_address(
                self.oracle_controller_address,
                self.client,
            )

        if self.account_contract is None:
            self.account_contract = await Contract.from_address(
                self.account_contract_address, self.client
            )

    async def get_balance(self):
        return await self.account_client.get_balance()

    async def send_transaction(self, to_contract, selector_name, calldata):
        selector = get_selector_from_name(selector_name)
        return await self.send_transactions([Call(to_contract, selector, calldata)])

    async def send_transactions(self, calls):
        return hex((await self.account_client.execute(calls, auto_estimate=True)).hash)



import warnings


def str_to_felt(text):
    if text.lower() != text:
        warnings.warn(
            "Converting string to felt that has uppercase characters. Converting to lowercase."
        )
        text = text.lower()
    b_text = bytes(text, "utf-8")
    return int.from_bytes(b_text, "big")


def felt_to_str(felt):
    num_bytes = (felt.bit_length() + 7) // 8
    bytes = felt.to_bytes(num_bytes, "big")
    return bytes.decode("utf-8")


def currency_pair_to_key(quote, base):
    return f"{quote}/{base}".lower()


def pprint_entry(entry):
    print(
        f"Entry: key={felt_to_str(entry.key)}, value={entry.value}, timestamp={entry.timestamp}, source={felt_to_str(entry.source)}, publisher={felt_to_str(entry.publisher)}"
    )


def key_for_asset(asset):
    key = asset["key"] if "key" in asset else currency_pair_to_key(*asset["pair"])

    return key



import os

from starknet_py.net.networks import TESTNET

ADMIN_ADDRESS = 0x0704CC0F2749637A0345D108AC9CD597BB33CCF7F477978D52E236830812CD98
PUBLISHER_REGISTRY_ADDRESS = (
    0x0743E8140A56D5EE9ED08EB77A92BCBCF8257DA34AB2A2EE93110709E61AB11A
)
ORACLE_CONTROLLER_ADDRESS = (
    0x012FADD18EC1A23A160CC46981400160FBF4A7A5EED156C4669E39807265BCD4
)

NETWORK = TESTNET
DEFAULT_AGGREGATION_MODE = 0

if os.environ.get("__EMPIRIC_STAGING_ENV__") == "TRUE":
    print("Warning: Communicating with staging contracts, not production")
    ORACLE_CONTROLLER_ADDRESS = (
        0x00225A37DE623DBD4D2287DDED4E0CB0EB4A5D7D9051D0E89A1321D4BCF9FDB2
    )
    PUBLISHER_REGISTRY_ADDRESS = (
        0x051949605AB53FCC2C0ADC1D53A72DD0FBCBF83E52399A8B05552F675B1DB4E9
    )



from collections import namedtuple

from empiric.core.utils import str_to_felt

Entry = namedtuple("Entry", ["key", "value", "timestamp", "source", "publisher"])


def serialize_entry(entry):
    return [entry.key, entry.value, entry.timestamp, entry.source, entry.publisher]


def serialize_entries(entries):
    expanded = [serialize_entry(entry) for entry in entries]
    flattened = [x for entry in expanded for x in entry]
    return [len(entries)] + flattened


def construct_entry(key, value, timestamp, source, publisher):
    if type(key) == str:
        key = str_to_felt(key)

    if type(publisher) == str:
        publisher = str_to_felt(publisher)

    if type(source) == str:
        source = str_to_felt(source)

    return Entry(
        key=key,
        value=value,
        timestamp=timestamp,
        source=source,
        publisher=publisher,
    )



from empiric.core.utils import key_for_asset

EMPIRIC_ALL_ASSETS = [
    {"type": "SPOT", "pair": ("BTC", "USD"), "decimals": 18},
    {"type": "SPOT", "pair": ("BTC", "EUR"), "decimals": 18},
    {"type": "SPOT", "pair": ("ETH", "USD"), "decimals": 18},
    {"type": "SPOT", "pair": ("SOL", "USD"), "decimals": 18},
    {"type": "SPOT", "pair": ("AVAX", "USD"), "decimals": 18},
    {"type": "SPOT", "pair": ("DOGE", "USD"), "decimals": 18},
    {"type": "SPOT", "pair": ("SHIB", "USD"), "decimals": 18},
    {"type": "SPOT", "pair": ("TEMP", "USD"), "decimals": 18},
    {"type": "SPOT", "pair": ("DAI", "USD"), "decimals": 18},
    {"type": "SPOT", "pair": ("USDT", "USD"), "decimals": 18},
    {"type": "SPOT", "pair": ("USDC", "USD"), "decimals": 18},
    {"type": "SPOT", "pair": ("TUSD", "USD"), "decimals": 18},
    {"type": "SPOT", "pair": ("BUSD", "USD"), "decimals": 18},
    {"type": "SPOT", "pair": ("ETH", "MXN"), "decimals": 18},
    {"type": "SPOT", "pair": ("BNB", "USD"), "decimals": 18},
    {"type": "SPOT", "pair": ("ADA", "USD"), "decimals": 18},
    {"type": "SPOT", "pair": ("XRP", "USD"), "decimals": 18},
    {"type": "SPOT", "pair": ("MATIC", "USD"), "decimals": 18},
    {"type": "SPOT", "pair": ("AAVE", "USD"), "decimals": 18},
    {"type": "FUTURE", "pair": ("BTC", "USD"), "decimals": 18},
    {"type": "FUTURE", "pair": ("ETH", "USD"), "decimals": 18},
    {
        "type": "ONCHAIN",
        "source": "AAVE",
        "key": "aave-on-borrow",
        "detail": {
            "asset_name": "USD Coin",
            "asset_address": "0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb480xb53c1a33016b2dc2ff3653530bff1848a515c8c5",
            "metric": "variableBorrowRate",
        },
        "decimals": 18,
    },
]

_EMPIRIC_ASSET_BY_KEY = {
    key_for_asset(asset): asset
    for asset in EMPIRIC_ALL_ASSETS
    if asset["type"] == "SPOT"
}


def get_spot_asset_spec_for_key(key):
    return _EMPIRIC_ASSET_BY_KEY[key]



import base64
import datetime
import hmac
import os
from hashlib import sha256

import requests
from empiric.core.entry import construct_entry
from empiric.core.utils import currency_pair_to_key


def fetch_coinbase(assets, publisher):
    source = "coinbase"
    COINBASE_API_SECRET = os.environ.get("COINBASE_API_SECRET")
    COINBASE_API_KEY = os.environ.get("COINBASE_API_KEY")
    COINBASE_API_PASSPHRASE = os.environ.get("COINBASE_API_PASSPHRASE")
    URL = "https://api.exchange.coinbase.com"
    REQUEST_PATH = "/oracle"
    METHOD = "GET"

    entries = []

    for asset in assets:
        if asset["type"] != "SPOT":
            print(f"Skipping Coinbase for non-spot asset {asset}")
            continue

        pair = asset["pair"]
        key = currency_pair_to_key(*pair)

        if pair[1] != "USD":
            print(f"Unable to fetch Coinbase price for non-USD denomination {pair[1]}")
            continue

        request_timestamp = str(
            int(
                datetime.datetime.now(datetime.timezone.utc)
                .replace(tzinfo=datetime.timezone.utc)
                .timestamp()
            )
        )

        signature = hmac.new(
            base64.b64decode(COINBASE_API_SECRET),
            (request_timestamp + METHOD + REQUEST_PATH).encode("ascii"),
            sha256,
        )

        headers = {
            "Accept": "application/json",
            "CB-ACCESS-KEY": COINBASE_API_KEY,
            "CB-ACCESS-SIGN": base64.b64encode(signature.digest()),
            "CB-ACCESS-TIMESTAMP": request_timestamp,
            "CB-ACCESS-PASSPHRASE": COINBASE_API_PASSPHRASE,
        }

        response = requests.request(
            METHOD, URL + REQUEST_PATH, headers=headers, timeout=10
        )

        response.raise_for_status()
        result = response.json()
        if pair[0] in result["prices"]:
            price = float(result["prices"][pair[0]])
            price_int = int(price * (10 ** asset["decimals"]))

            timestamp = int(result["timestamp"])

            print(f"Fetched price {price} for {key} from Coinbase")

            entries.append(
                construct_entry(
                    key=key,
                    value=price_int,
                    timestamp=timestamp,
                    source=source,
                    publisher=publisher,
                )
            )
        else:
            print(f"No entry found for {key} from Coinbase")

    return entries



import time

import requests
from empiric.core.entry import construct_entry
from empiric.core.utils import currency_pair_to_key


def fetch_cryptowatch(assets, publisher):
    sources = [
        "cryptowatch-coinbase-pro",
        "cryptowatch-kraken",
        "cryptowatch-binance",
        "cryptowatch-bitfinex",
    ]

    response = requests.get("https://api.cryptowat.ch/markets/prices")
    result = response.json()["result"]

    entries = []

    for source in sources:

        cryptowatch_source = source.split("-", 1)[1]
        source_results = {
            k.split(":")[2]: v
            for k, v in result.items()
            if k.startswith(f"market:{cryptowatch_source}")
        }

        for asset in assets:
            if asset["type"] != "SPOT":
                print(f"Skipping Cryptowatch for non-spot asset {asset}")
                continue

            pair = asset["pair"]
            key = currency_pair_to_key(*pair)

            try:
                price = source_results["".join(pair).lower()]
            except KeyError:
                print(f"No entry found for {key} from Cryptowatch-{cryptowatch_source}")
                continue

            timestamp = int(time.time())
            price_int = int(price * (10 ** asset["decimals"]))

            print(
                f"Fetched price {price} for {'/'.join(pair)} from Cryptowatch-{cryptowatch_source}"
            )

            entries.append(
                construct_entry(
                    key=key,
                    value=price_int,
                    timestamp=timestamp,
                    source=source,
                    publisher=publisher,
                )
            )

    return entries
